////////////////////////////////////////////////////////////////////////
// Class:       ActivityTrigger
// Module Type: filter
// File:        ActivityTrigger_module.cc
//
// Generated at Wed Sep 12 16:15:01 2012 by Nova DDT using artmod
// from art v1_01_01.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Persistency/Common/Assns.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/TriggerDecision.h"
#include "DDTBaseDataProducts/GroupedHitList.h"
#include "DDTUtilities/AssociationUtil.h"

#include <string>
#include <algorithm>
#include <unordered_set>

#define MAX_PLANE 999999
#define MIN_PLANE -1

namespace novaddt {
  class ActivityTrigger;
}

class novaddt::ActivityTrigger : public art::EDFilter {

public:
  explicit ActivityTrigger(fhicl::ParameterSet const & p);
  virtual ~ActivityTrigger();

  virtual bool filter(art::Event & e);

private:

  std::string  _hitsModuleLabel;   ///< label of module making the HitList
  std::string  _hitsInstanceLabel; ///< instance label making the HitList
  unsigned int _minHits;           ///< Minimum number of hits in the slice
  unsigned int _minPlanes;         ///< Minimum planes (unique) the hits are spread across
           int _minExtent;         ///< Minimum delta between lowest and highest z plane
  unsigned int _prescale;          ///< holds prescaling variable from fcl file
  
  unsigned int _triggerCounts = 0;
  unsigned int _sliceCounts = 0;
  unsigned int _evtCounts = 0;
};

//---------------------------------------------------------------------    
novaddt::ActivityTrigger::ActivityTrigger(fhicl::ParameterSet const & p)
  : _hitsModuleLabel  (p.get<std::string >("HitModuleLabel"  ))
  , _hitsInstanceLabel(p.get<std::string >("HitInstanceLabel"))
  , _minHits          (p.get<unsigned int>("MinHits"         ))
  , _minPlanes        (p.get<unsigned int>("MinPlanes"       ))
  , _minExtent        (p.get<unsigned int>("MinExtent"       ))
  , _prescale         (p.get<unsigned int>("prescale"        ))
{
  produces< std::vector<novaddt::TriggerDecision> >();
  //produces< art::Assns<novaddt::TriggerDecision, novaddt::HitList> >();
}

//---------------------------------------------------------------------    
novaddt::ActivityTrigger::~ActivityTrigger()
{
  // Clean up dynamic memory and other resources here.
}

//---------------------------------------------------------------------    
bool novaddt::ActivityTrigger::filter(art::Event & e)
{
  //std::cout << "--- ActivityTrigger filter\n";
  ++_evtCounts;

  // Retrieve the (grouped) hit list for this event
  art::Handle<novaddt::GroupedHitList<novaddt::TDC> > hits;
  e.getByLabel(_hitsModuleLabel, _hitsInstanceLabel, hits);
  //std::cout << "\tgot "     << hits->n_groups()
  //          << " groups.\n";

  // Allocate the vector for the trigger decisions that we may make (and write to the event)
  std::unique_ptr<std::vector<novaddt::TriggerDecision>> td(new std::vector<novaddt::TriggerDecision>);

  //std::unique_ptr<art::Assns<novaddt::TriggerDecision, novaddt::HitList> > assn(new art::Assns<novaddt::TriggerDecision, novaddt::HitList >);

  /**************************************************************************
   * Activity Trigger Algorithm                                             *
   *                                                                        *
   * Essentially all we do is count the number of hits we have in slice     *
   * if it is over a given threshold then we write it out.                  *
   *                                                                        *
   * Overlaps with the beam (numi) trigger are handled via an inhibit at    *
   * the global trigger level                                               *
   *                                                                        *
   * However there are a couple of additional cuts we put on the data,      *
   * first in addition to having at least N hits we require that they       *
   * extend over at least M planes and represent at least P unique planes   *
   *                                                                        *
   * This allows the trigger to be tuned in the following manners           *
   * Increase: N = more hits (more activity, larger showers                 *
   * Increase: M and P = more "horizontal" (z-aligned) activity             *
   * Increase: N, decrease M and P = more "vertical" (x/y-aligned activity) *
   **************************************************************************/

  int planeLow;
  int planeHigh;

  std::unordered_set<int> planeList;

  // Loop over the slices in the event
  for(unsigned int i=0; i<hits->n_groups(); ++i){

    // Get the hits for the current slice
    novaddt::HitList theSliceHits = hits->get(i);

    /*
    if(theSliceHits.size() > 1) {
      std::cout << "\tgroup[" << i
              << "] has "   << theSliceHits.size()
              << " hits "   << "out of a min: "
     	        << _minHits   << "\n";
    }
    */

    ++_sliceCounts;

    // Test to see how many hits are in the slice
    if (theSliceHits.size() > _minHits){
      // If there are enough hits find the delta between
      // lowest and highest plane number
      planeLow=MAX_PLANE;
      planeHigh=MIN_PLANE;
      planeList.clear();
      // Find the high and low planes
      for(unsigned int j=0; j < theSliceHits.size(); ++j){
	auto thePlane = theSliceHits.at(j).Plane().val;
	if(thePlane > planeHigh){ planeHigh = thePlane;}
	if(thePlane < planeLow){  planeLow = thePlane; }
	// Push the plane into the list of unique planes
	planeList.insert(thePlane);
      } // endfor j
      
      if( (planeHigh - planeLow) > _minExtent){
	// If the hits are spread over the min extent then see that 
	// there are at least minPlane unique planes
	if(planeList.size() > _minPlanes){
	  /******************************************************
	   * EVENTS THAT MAKE IT HERE SHOULD GENERATE A TRIGGER *
	   ******************************************************/

	  // Internal prescale 
	  ++_triggerCounts;
	  if (_triggerCounts % _prescale == _prescale - 1){
	    
	    // Store a trigger decision for this slice
	    td->emplace_back(theSliceHits.front().TDC().val, 
			     theSliceHits.back().TDC().val - theSliceHits.front().TDC().val, 
			     daqdataformats::trigID::TRIG_ID_DATA_ACTIVITY_1,
			     _prescale);

	  } // end of issue trigger decision	  
	} // endif gt minPlanes
      } // endif gt minExtent
    } // end of check on hit list size
  } // end loop on hit groups

  bool goodTrigger = (td->size() > 0);

  e.put(std::move(td));
  //e.put(std::move(assn));

  //std::cout << "\nA minimum of " << _minHits << " hits yields:" << "\n";
  //std::cout << "Frequency: " << _triggerCounts*1./(_evtCounts*5.e-3) << " pm " // Fraction of passing per event. One event lasts 5 ms 
  //          << sqrt(_triggerCounts*1.)/(_evtCounts*5.e-3) << " Hz" << std::endl; // Error, assuming poissonian behaviour

  return goodTrigger;
}

DEFINE_ART_MODULE(novaddt::ActivityTrigger)
