////////////////////////////////////////////////////////////////////////
// Class:       UpMuTrigger
// Module Type: filter
// File:        UpMuTrigger_module.cc
//
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Persistency/Common/Assns.h"
#include "art/Framework/Principal/Handle.h"	
#include "art/Framework/Principal/Event.h"
#include "DDTUtilities/AssociationUtil.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindOneP.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/Boundary.h"
#include "DDTBaseDataProducts/BoundaryList.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/Track.h"
#include "DDTBaseDataProducts/Track3D.h"
#include "DDTBaseDataProducts/TriggerDecision.h"
#include "DDTBaseDataProducts/Track.h"	

#include <TGraphErrors.h>
#include <TF1.h>
#include <string>
#include <vector>
#include <algorithm>
#include <numeric>
#include <boost/math/special_functions/gamma.hpp>
#include <boost/math/special_functions.hpp>

namespace novaddt {
  class UpMuTrigger;
}

class novaddt::UpMuTrigger : public art::EDFilter {
public:
  explicit UpMuTrigger(fhicl::ParameterSet const & p);
  virtual ~UpMuTrigger();

  virtual bool filter(art::Event & e) override;

private:
  unsigned _prescale;
  std::string _slicelabel;
  std::string _sliceinstance;
  std::string _tracklabel;
  std::string _trackinstance;
  std::string _trackToSlice;

  std::string _detector;

  unsigned _trigger_counts, _after_prescale;

  float c0x, c0y, p0, cw, pw; // initialized later based on detector
  float _TrackLen; 
  unsigned _TrackHitsXY;
  unsigned _TrackHitsX;
  unsigned _TrackHitsY; 
  int _dX; 
  int _dY; 
  int _dZ; 
  float _Chi2;
  float _LLR; 
  float _R2X; 
  float _R2Y; 

  float GetT(double tdc, float dist_to_readout);
  float GetRes(int ADC);
  void GetXYZ(DAQHit const& daqhit, TVector3 start, TVector3 end, double *xyzd);

  void LinFit(const std::vector<float>& x, const std::vector<float>& y, float *fitpar);
  void LinFit(const std::vector<float>& x, const std::vector<float>& y, const std::vector<float>& ye, float *fitpar);
  void LinFitLLR(std::vector<float>& x_hit, std::vector<float>& y_hit, std::vector<float>& y_error, 
		 float& slope, float& chi2, float& P_up, float& P_dn);

  double GetPigtail(DAQHit const& daqhit) const;
  float  getDCMOff(uint16_t cell, uint8_t view);
};

novaddt::UpMuTrigger::UpMuTrigger(fhicl::ParameterSet const & p)
  : _prescale       (p.get<unsigned>("prescale"))
  , _slicelabel     (p.get<std::string>("slice_label"))
  , _sliceinstance  (p.get<std::string>("slice_instance"))
  , _tracklabel     (p.get<std::string>("track_label"))
  , _trackinstance  (p.get<std::string>("track_instance"))
  , _trackToSlice   (p.get<std::string>("TrackToSliceInstanceLabel"))
  , _detector       (p.get<std::string>("detector"))
				 
  , _TrackLen       (p.get<float>("TrackLen"))
  , _TrackHitsXY    (p.get<unsigned>("TrackHitsXY"))
  , _TrackHitsX     (p.get<unsigned>("TrackHitsX"))
  , _TrackHitsY     (p.get<unsigned>("TrackHitsY"))
  , _dX             (p.get<int>("dX"))
  , _dY             (p.get<int>("dY"))
  , _dZ             (p.get<int>("dZ"))
  , _Chi2           (p.get<float>("Chi2"))
  , _LLR            (p.get<float>("LLR"))
  , _R2X            (p.get<float>("R2X"))
  , _R2Y            (p.get<float>("R2Y"))

{
  assert(_detector == "ndos" || _detector == "fd");

  //initialize cell and plane locations
  if (_detector == "ndos") {
    c0x = -119.5;
    c0y = -199.5;
    p0  = 4.57;
    cw = 3.97;
    pw = 6.654;
  } else if (_detector == "fd") {
    c0x = -759.5;
    c0y = -759.5;
    p0  = 4.57;
    cw = 3.97;
    pw = 6.654;
  }

  _trigger_counts=0;
  _after_prescale=0;

  produces<std::vector<TriggerDecision>>();
}

novaddt::UpMuTrigger::~UpMuTrigger()
{
  // Clean up dynamic memory and other resources here.
}

bool novaddt::UpMuTrigger::filter(art::Event & e)	
{
   std::unique_ptr<std::vector<TriggerDecision>> 
     trigger_decisions(new std::vector<TriggerDecision>);
   bool result = false;

   art::Handle<std::vector<novaddt::HitList> > slices;
   e.getByLabel(_slicelabel, _sliceinstance, slices);

   art::Handle<std::vector<novaddt::Track> > tracks;
   e.getByLabel("track", "", tracks);

   // get the hit list for each track
   art::FindOneP<novaddt::HitList> fohl(tracks, e, _tracklabel);
   // get the tracks for each hit list
   const std::string the_slices = _tracklabel+":"+_trackToSlice;
   
   art::FindManyP<novaddt::Track3D> find_tracks(slices, e, the_slices);
   assert(find_tracks.isValid());

   // Loop through the slices
    for(size_t i_slice = 0; i_slice < slices->size(); ++i_slice){

      novaddt::HitList hit_list = slices->at(i_slice);
      
      // find the tracks for this hit list
      std::vector< art::Ptr<novaddt::Track3D> > tracks = find_tracks.at(i_slice);    
      if(tracks.size() == 0) continue;
      // get the hit list for each track
      art::FindOneP<novaddt::HitList> fohl(tracks, e, _tracklabel);
      assert(fohl.isValid());
      assert(fohl.size() == tracks.size());

      for(size_t i_track = 0; i_track < tracks.size(); ++i_track){
	art::Ptr<novaddt::Track3D> track = tracks[i_track];
	// get the hit list for this track
	art::Ptr<novaddt::HitList> this_hit_list = fohl.at(i_track);

	int StartX = track->Start().X();
	int StartY = track->Start().Y();
	int StartZ = track->Start().Z();

	int EndX = track->End().X();
	int EndY = track->End().Y();
	int EndZ = track->End().Z();
	if(StartY > EndY){
	  std::swap(StartX, EndX);
	  std::swap(StartY, EndY);
	  std::swap(StartZ, EndZ);
	}

	TVector3 length;
	length.SetXYZ(cw*(track->End().X() - track->Start().X()),
		      cw*(track->End().Y() - track->Start().Y()),
		      pw*(track->End().Z() - track->Start().Z())
		      );

	double this_length = length.Mag();
	float exp_trav_time = this_length/29.97;   // c=29.97 cm/ns; 1/64E6=15.625 ns


	if (this_length < _TrackLen)       continue; // Select tracks longer than _TrackLen.
	if (TMath::Abs(EndX-StartX) < _dX) continue; // Select tracks longer than _dX cells in X.
	if (TMath::Abs(EndY-StartY) < _dY) continue; // Select tracks longer than _dY cells in Y.
	if (TMath::Abs(EndZ-StartZ) < _dZ) continue; // Select tracks longer than _dZ planes in Z.
	if (!track->Is3D())                continue; // Select only 3D tracks

	uint64_t T0 = 0;
	std::vector<float> x_hit;
	std::vector<float> y_hit;
	std::vector<float> zy_hit;
	std::vector<float> zx_hit;

	// Get T0 that corresponds to the minimum cell number in Y-view
	for(size_t i_hit = 0; i_hit < this_hit_list->size(); ++i_hit){
	  uint16_t iC = this_hit_list->at(i_hit).Cell().val;
	  uint16_t iP = this_hit_list->at(i_hit).Plane().val;
	  uint8_t view = this_hit_list->at(i_hit).View().val;
	  uint64_t iTDC = this_hit_list->at(i_hit).TDC().val;	
	  
	  if(view == daqchannelmap::X_VIEW){
	    x_hit.push_back(iC);
	    zx_hit.push_back(iP);
	  }

	  if(view == daqchannelmap::Y_VIEW){
	    if(iC == StartY)
	      T0 = iTDC;
	    y_hit.push_back(iC);
	    zy_hit.push_back(iP);
	  }
	  
	}


	// Linear fit of the hits in XZ/YZ views. Drop tracks that has bad linear fit


	float fitpar[3];
	LinFit(x_hit, zx_hit, fitpar);
	float r2x = fitpar[2];
	LinFit(y_hit, zy_hit, fitpar);
	float r2y = fitpar[2];

	if(r2x < _R2X) continue;
	if(r2y < _R2Y) continue;

	// Vectors of expected and measured times for TGraph 
	std::vector<float> eT_vecxy;
	std::vector<float> mT_vecxy;

	// errors
	std::vector<float> mT_vecxy_e;
	//std::vector<floatt> eT_vecxy_e;

	for(size_t i_hit = 0; i_hit < this_hit_list->size(); ++i_hit){
	  uint16_t iC = this_hit_list->at(i_hit).Cell().val;
	  //uint16_t iP = this_hit_list->at(i_hit).Plane().val;
	  int hit_adc = this_hit_list->at(i_hit).ADC().val;

	  if (hit_adc < 50) continue;


	  // Calculate time of the hit
	  DAQHit const& dqhit = this_hit_list->at(i_hit);
	  double pos[4];
	  TVector3 start = track->Start();
	  TVector3 end = track->End();
	  GetXYZ(dqhit, start, end, pos);
	  double pigtail = GetPigtail(dqhit);

	  const uint64_t remain = 10000;
	  double  meas_time = ((double)(this_hit_list->at(i_hit).TDC().val%remain) - (double)(T0%remain));
	  meas_time = meas_time + 0.01*static_cast<double>((this_hit_list->at(i_hit).TDC().frac));


	  float tns = GetT(meas_time, pos[3] + pigtail);
	  // Subtract DCM offsects
 	  float DCM_off = getDCMOff(iC, this_hit_list->at(i_hit).View().val);
 	  tns = tns - DCM_off;

	  float exp_time = -999;
	  std::string h_view = "y";	   

	  if( this_hit_list->at(i_hit).View().val == daqchannelmap::X_VIEW){
	    h_view = "x";
	    exp_time = exp_trav_time*(iC-StartX)/(EndX-StartX);
	  }
	  else	    	    
	    exp_time = exp_trav_time*(iC-StartY)/(EndY-StartY);

	  assert(pos[3]>=0);	  

	  eT_vecxy.push_back(exp_time);
	  mT_vecxy.push_back(tns);
	  mT_vecxy_e.push_back(GetRes(hit_adc));

	} // end of loop on hits

	unsigned nhit = eT_vecxy.size();

	// Remove tracks with less than fTrackHitsXY hits in X+Y views
	if(nhit < _TrackHitsXY) continue;

	float slope_xy, Chi2XY, P_up_xy, P_dn_xy;
	LinFitLLR(eT_vecxy, mT_vecxy, mT_vecxy_e, slope_xy, Chi2XY, P_up_xy, P_dn_xy);
	double LLR = log(P_up_xy/P_dn_xy);

	// Remove tracks with LLR smaller than input in the config fcl file
	if(LLR < _LLR) continue;
	// Remove tracks with Chi2 fit values higher than input in the config fcl file
	if(Chi2XY > _Chi2) continue;
       
	_trigger_counts++;
	if (_trigger_counts%_prescale == _prescale-1) {
	  _after_prescale++;
	  // get the min and max TDC from the parent slice 
	  // and use these to define the trigger window
	  // safest not to assume this is time sorted
	  uint64_t min_tdc = hit_list.front().TDC().val;
	  uint64_t max_tdc = hit_list.back().TDC().val;
	  for(size_t i_hit = 0; i_hit < hit_list.size(); ++i_hit){
	    if (hit_list[i_hit].TDC().val < min_tdc) min_tdc = hit_list[i_hit].TDC().val;
	    if (hit_list[i_hit].TDC().val > max_tdc) max_tdc = hit_list[i_hit].TDC().val;
	  } // end of loop on hits

	  trigger_decisions->emplace_back
	    (min_tdc, max_tdc - min_tdc, daqdataformats::trigID::TRIG_ID_DATA_UPMU, _prescale);

	  result = true;
	  break; // don't need to loop any more on this slice
	} // end if (prescale)
	
      } // end of loop on tracks

    } // end of loop on slices

  e.put(std::move(trigger_decisions));
  return result;

} // end Filter()

void novaddt::UpMuTrigger::GetXYZ(DAQHit const& daqhit, TVector3 start, TVector3 end, double *xyzd) {
  UInt_t plane = daqhit.Plane().val;
  UInt_t cell  = daqhit.Cell().val;

  // The parameters are taken from the fit of plane vs Z
  float x = c0x + cw*cell;
  float y = x;
  float z = p0 + pw*plane;
  float d = -999.9;

  float x0 = c0x + cw*start.X(); float x1 = c0x + cw*end.X();  
  float y0 = c0y + cw*start.Y(); float y1 = c0y + cw*end.Y();    
  float z0 = p0 + pw*start.Z(); float z1 = p0 + pw*end.Z();

  if(daqhit.View().val == daqchannelmap::X_VIEW){
    y = (y1 - y0)/(z1 - z0);
    y = y*(z - z0);
    y = y + y0;
    d = 800 - y;
  }
  if(daqhit.View().val == daqchannelmap::Y_VIEW){
    x = (x1 - x0)/(z1 - z0);
    x = x*(z - z0);
    x = x + x0;
    d = 800 - x;
  }

  if(d<0) d=0;
  xyzd[0] = x;
  xyzd[1] = y;
  xyzd[2] = z;
  xyzd[3] = d;
}

float novaddt::UpMuTrigger::GetT(double tdc, float dist_to_readout) {
  // This needs to be calibrated and put in a DB table
  float speedOfFiberTransport = 15.3; // cm/ns, "first principles" calib.
  float TDC_to_ns = 15.625; // 1/64E6=15.625 ns
  
  // Differs from c/n due to zigzag
  // paths down fiber.  But, this
  // value may be way off (10%? 30%?).
  return (tdc*TDC_to_ns - dist_to_readout/speedOfFiberTransport);
}



float novaddt::UpMuTrigger::GetRes(int ADC) {

  //The fit function from Evan: f(x) = p0/(p1+x^p2) + p3;
  float p0 = 70267.4;
  float p1 = 677.305;
  float p2 = 1.48702;
  float p3 = 8.46997;
  
  float res = p0/(p1 + pow( (float)ADC, p2) ) + p3;
  return res;

}


double novaddt::UpMuTrigger::GetPigtail(DAQHit const& daqhit) const{
    if (_detector == "fd") {
      UInt_t cellid  = daqhit.Cell().val;
      //NOTE: Do we think 32 cells per module can ever change
      //This number can be found in DAQChannelMap but did not want to add that
      //dependancy to geometry unless needed.
      const int kCellsPerModule = 32;
      cellid = cellid % kCellsPerModule;
      // In vertical planes, high cell numbers have longer fibres.
      // In horizontal planes it's the opposite, so flip it round.
      if(daqhit.View().val == daqchannelmap::X_VIEW)
        cellid = kCellsPerModule-cellid-1;

      // This should really never happen, but just to be safe...
      if(cellid < 0 || cellid >= kCellsPerModule) return 100;
      // Email from Tom Chase 2011-04-29
      // NB: this isn't just a ~3.8cm change per cell. Every 8 cells something
      // different happens.
      const double kPigtails[kCellsPerModule] = {
        34.5738, 38.4379, 42.3020, 46.1660, 50.0301, 53.8942, 57.7582, 61.6223,
        64.7504, 68.6144, 72.4785, 76.3426, 80.2067, 84.0707, 87.9348, 91.0790,
        95.3301, 99.1941, 103.058, 106.922, 110.786, 114.650, 118.514, 122.379,
        125.507, 129.371, 133.235, 137.099, 140.963, 144.827, 148.691, 150.751
      };
      return kPigtails[cellid];
    }
    else
      return 0; //to be updated as necessary
}


void novaddt::UpMuTrigger::LinFit(const std::vector<float>& x_val, const std::vector<float>& y_val, float *fitpar){
  // http://en.wikipedia.org/wiki/Simple_linear_regression
  const auto n    = x_val.size();
  const auto x  = (std::accumulate(x_val.begin(), x_val.end(), 0.0))/n;                       // <x>
  const auto y  = (std::accumulate(y_val.begin(), y_val.end(), 0.0))/n;                       // <y>
  const auto xx = (std::inner_product(x_val.begin(), x_val.end(), x_val.begin(), 0.0))/n;     // <xx>
  const auto yy = (std::inner_product(y_val.begin(), y_val.end(), y_val.begin(), 0.0))/n;     // <yy>
  const auto xy = (std::inner_product(x_val.begin(), x_val.end(), y_val.begin(), 0.0))/n;     // <xy>

  const auto b    = (xy - x*y)/(xx - x*x);                                                    // slope
  const auto a    = y - b*x;                                                                  // intercept
  const auto r    = (xy - x*y)/sqrt((xx - x*x)*(yy - y*y));                                   // Rxy - coeffcient of determination
  fitpar[0] = a;
  fitpar[1] = b;
  fitpar[2] = r*r;

}

void novaddt::UpMuTrigger::LinFit(const std::vector<float>& x_val, const std::vector<float>& y_val, const std::vector<float>& y_err, float *fitpar){
  // http://en.wikipedia.org/wiki/Simple_linear_regression
  int n    = x_val.size();
  float x = 0; 
  float y = 0; 
  float xx = 0;
  float yy = 0;
  float xy = 0;
  float ee = 0;

  for ( int i=0; i<n; i++ ){

    x = x + x_val[i]/y_err[i]/y_err[i];
    y = y + y_val[i]/y_err[i]/y_err[i];

    xx = xx + x_val[i]*x_val[i]/y_err[i]/y_err[i];
    yy = yy + y_val[i]*y_val[i]/y_err[i]/y_err[i];
    xy = xy + x_val[i]*y_val[i]/y_err[i]/y_err[i];
    ee = ee + 1./y_err[i]/y_err[i];
  }

  const auto b    = (xy*ee - x*y)/(xx*ee - x*x);                                              // slope
  const auto a    = (xy - b*xx)/x;                                                            // intercept
  const auto r    = (xy - x*y)/sqrt((xx - x*x)*(yy - y*y));                                   // Rxy - coeffcient of determination
  fitpar[0] = a;
  fitpar[1] = b;
  fitpar[2] = r*r;


}

void novaddt::UpMuTrigger::LinFitLLR(std::vector<float>& x_hit, std::vector<float>& y_hit, std::vector<float>& y_error, 
				 float& slope, float& chi2, float& P_up, float& P_dn) {

  int n = x_hit.size();

  float fitpar[3];
  LinFit(x_hit, y_hit, y_error, fitpar);	
  float a = fitpar[0];
  float b = fitpar[1];


  int totAllowOutlier = (int)(0.05*n);
  const float sigma_cut = 3;
  int iOutlier = 0;
  std::vector<float> x_filt, y_filt, ye_filt, x_filt_e;
  for (int i = 0; i < n; i++) {
    float y_fit = a + b*x_hit[i];
    if ( fabs(y_hit[i] - y_fit) < sigma_cut*y_error[i] || iOutlier>totAllowOutlier) 
      { // > 5 * sigma. Accept 10% of 5sigma outliers. These numbers should be optimized.
	x_filt.push_back(x_hit[i]);
	y_filt.push_back(y_hit[i]);
	ye_filt.push_back(y_error[i]);
	x_filt_e.push_back(0);
      }    
    else
      iOutlier++;

  }

  LinFit(x_filt, y_filt, ye_filt, fitpar);	
  a = fitpar[0];
  b = fitpar[1];
  n = x_filt.size();
  if (n < 5){
    slope = 0;
    chi2  = 999;
    P_dn   = 1e-30;
    P_dn   = 1e-30;
    return;
  }


  slope = fitpar[1];
  
  // chi2
  chi2=0.0;
  for (int i=0; i<n; i++) {
    float y_expected = a + b*x_filt.at(i);
    chi2+=pow((y_filt.at(i)-y_expected) / ye_filt.at(i), 2.0);
  }

  chi2/=(float)(n-2);      // NDF = N points - 2 parameters in the fit


  // Calculate up/down intercepts
  float one_over_ee = 0.0; 
  float x_over_ee   = 0.0;
  float y_over_ee   = 0.0;
  float xx_over_ee  = 0.0;
  float xy_over_ee  = 0.0;

  for (int i=0; i<n; i++) {
    float e = ye_filt.at(i);
    one_over_ee+=1.0/e/e;
    xy_over_ee+=x_filt.at(i)*y_filt.at(i)/e/e;
    xx_over_ee+=x_filt.at(i)*x_filt.at(i)/e/e;
    x_over_ee+=x_filt.at(i)/e/e;
    y_over_ee+=y_filt.at(i)/e/e;
  }

  // up/down intercepts defined below
  double up_inter   = (y_over_ee-x_over_ee)/one_over_ee;
  double down_inter = (y_over_ee+x_over_ee)/one_over_ee;


  // Calculate up/down chi2  			
  float up_chi2=0.0, down_chi2=0.0;
  for (int i=0;i<n;i++) {

    float e = ye_filt.at(i);
    float up_expected   = up_inter + x_filt.at(i);
    float down_expected = down_inter - x_filt.at(i);
    up_chi2   += pow((y_filt.at(i)-up_expected) / e, 2.0);
    down_chi2 += pow((y_filt.at(i)-down_expected) / e, 2.0);
  }
  
  // if statement is here for safety to prevent gamma_q failing

    float prob_up   = boost::math::gamma_q((float)(n-2)/2.0,up_chi2/2.0);
    float prob_down = boost::math::gamma_q((float)(n-2)/2.0,down_chi2/2.0);

    // Use Root TMath if you prefer. Root and boost libraries give identical results.
    // float prob_up = TMath::Prob(up_chi2, n-2);
    // float prob_down = TMath::Prob(down_chi2, n-2);

    if (prob_up<1e-30) prob_up = 1e-30;
    if (prob_down<1e-30) prob_down = 1e-30;

    P_up = prob_up;
    P_dn = prob_down;

}

float novaddt::UpMuTrigger::getDCMOff(uint16_t cell, uint8_t view){
  if (_detector == "ndos") {
    if (view == daqchannelmap::Y_VIEW)
      if (cell > 31 && cell < 64) return -26;
      else if (cell > 63) return -40;
      else return 0;
    else // X_VIEW
      if (cell > 7 && cell < 32) return -10;
      else if (cell > 31) return 5;
      else return 0;
  }
  else if (_detector == "fd")
    return (cell/64)*40.;
  else return 0;
}

DEFINE_ART_MODULE(novaddt::UpMuTrigger)
